/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <assert.h>
#include "is_types.h"
#include "is_proto.h"

#ifndef lint
static	char	sccsid[] = "@(#)db_read.c	1.10	08:43:16	1/31/92";
#endif

extern	int	db_errno;

int
__db_read_curr (file, buffer)
file_t	*file;
char	*buffer;
{
	int	size;
	keyblk_t keyblk;
	off_t	block;
	sort_key_t *key;
	int	was_locked;
	long	ticket;

	if (! (was_locked = file->f_locked)) {
		if ((ticket = isam_get_ticket (file)) == -1)
			return -1;

		/*
		 * NEEDSWORK: This should cause the current record to
		 * be re-read and looked up in the database.  From
		 * that position the code could continue.
		 */

		if (ticket != file->f_ticket)
			goto fail;
	}
	key = &file->f_index->i_keys[file->f_key_index];
	size = key->k_key_size + sizeof (off_t);

	if (isam_read_data (file, file->f_current_record, buffer))
		goto fail;

	if (! was_locked)
		isam_end_ticket (file);

	return 0;

fail:
	if (! was_locked)
		isam_end_ticket (file);

	return -1;
}

int
__db_read_next (file, buffer)
file_t	*file;
char	*buffer;
{
	int	size;
	keyblk_t keyblk;
	off_t	block;
	sort_key_t *key;
	int	was_locked;
	long	ticket;

	if (! (was_locked = file->f_locked)) {
		if ((ticket = isam_get_ticket (file)) == -1)
			return -1;

		/*
		 * NEEDSWORK: This should cause the current record to
		 * be re-read and looked up in the database.  From
		 * that position the code could continue.
		 */

		if (ticket != file->f_ticket)
			goto fail;
	}
	key = &file->f_index->i_keys[file->f_key_index];
	size = key->k_key_size + sizeof (off_t);

	if ((block = isam_next (file, file->f_current_index)) <= 0)
		goto fail;

	if (isam_read_data (file, block, buffer))
		goto fail;

	if (! was_locked)
		isam_end_ticket (file);

	file->f_current_record = block;
	return 0;

fail:
	if (! was_locked)
		isam_end_ticket (file);

	return -1;
}

int
__db_read_prev (file, buffer)
file_t	*file;
char	*buffer;
{
	int	size;
	keyblk_t keyblk;
	off_t	block;
	sort_key_t *key;
	int	rc;
	int	was_locked;
	long	ticket;

	if (! (was_locked = file->f_locked)) {
		if ((ticket = isam_get_ticket (file)) == -1)
			return -1;

		/*
		 * NEEDSWORK: This should cause the current record to
		 * be re-read and looked up in the database.  From
		 * that position the code could continue.
		 */

		if (ticket != file->f_ticket)
			goto fail;
	}
	key = &file->f_index->i_keys[file->f_key_index];
	size = key->k_key_size + sizeof (off_t);

	if ((block = isam_previous (file, file->f_current_index)) <= 0)
		goto fail;

	if (isam_read_data (file, block, buffer))
		goto fail;

	if (! was_locked)
		isam_end_ticket (file);

	file->f_current_record = block;
	return 0;

fail:
	if (! was_locked)
		isam_end_ticket (file);

	return -1;
}

int
__db_read_record (file, index, data, match, buffer)
file_t	*file;
int	index;
VOIDP	*data;
int	match;
char	*buffer;
{
	off_t	record;
	int	was_locked;
	long	ticket;
	int	unique;

	if (! (was_locked = file->f_locked)) {
		if ((ticket = isam_get_ticket (file)) == -1)
			return -1;
	}

	/*
	 * Figure out how the index is to be treated.  Index 1 is almost
	 * always a unique index.  However, sometimes it isn't.  This is
	 * a terrible kludge, but it makes other things easier, so we
	 * do it ...
	 */

	if (index == 1) {
		if (file->f_index->i_keys->k_unique)
			unique = IS_UNIQUE;
		else
			unique = IS_NONUNIQUE;
	} else
		unique = IS_NONUNIQUE;

	if ((record = isam_lookup (file, index, data, match, unique)) == -1)
		goto fail;

	if (isam_read_data (file, record, buffer))
		goto fail;

	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	file->f_current_record = record;
	return 0;

fail:
	if (! was_locked)
		isam_end_ticket (file);

	return -1;
}

int
db_read_record (dbm, filenum, data, match)
dbm_t	*dbm;
int	filenum;
VOIDP	*data;
int	match;
{
	file_t	*file;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	if (file->f_index == 0)
		return (db_errno = DB_BAD_DB), -1;

	if (file->f_current_index <= 0 ||
			file->f_current_index > file->f_index->i_key_count)
		return (db_errno = DB_NO_INDEX), -1;

	if (match == IS_CURR)
		return __db_read_curr (file, file->f_data_buffer);
	else if (match == IS_NEXT)
		return __db_read_next (file, file->f_data_buffer);
	else if (match == IS_PREV)
		return __db_read_prev (file, file->f_data_buffer);
	else
		return __db_read_record (file, file->f_current_index,
			data, match, file->f_data_buffer);
}
