/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#ifndef lint
static	char	sccsid[] = "@(#)db_rename.c	1.4	09:17:11	3/30/92";
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include "is_types.h"
#include "is_proto.h"

#ifdef	BSD
#include <strings.h>
#define	strchr	index
#define	strrchr	rindex
#else
#include <string.h>
#ifndef	SYSIII
#include <memory.h>
#endif
#endif

#ifndef	STDLIB_H
char	*getenv();
#endif

extern	int	db_errno;

/*
 * db_rename_field - rename an existing database field
 */

int
db_rename_field (dbm, filenum, oldname, newname)
dbm_t	*dbm;
int	filenum;
char	*oldname;
char	*newname;
{
	file_t	*dict = &dbm->d_dictionary;
	file_t	*file;
	field_t	*field;
	struct	{
		char	filename[MAXFILENAME];
		char	fieldname[MAXFIELDNAME];
	} key;
	dict_t	record;
	int	i;

	/*
	 * Validate the file number.
	 */

	if (filenum <= 0 || filenum > dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	/*
	 * Validate the new field name.  It must not exist in selected
	 * file at present.
	 */

	for (i = 1;i < file->f_schema->s_field_count;i++)
		if (strncmp (file->f_schema->s_fields[i].f_name,
				newname, MAXFIELDNAME) == 0)
			return (db_errno = DB_FIELD_EXISTS), -1;

	/*
	 * Rewind the database dictionary
	 */

	dict->f_current_index = 1;
	if (isam_rewind (dict))
		return -1;

	/*
	 * Build the search key that will be used to locate the
	 * data dictionary entry for this particular field.
	 */

#ifdef	BSD
	bzero (&key, sizeof key);
#else
	memset ((char *) &key, 0, sizeof key);
#endif
	strncpy (key.filename, file->f_file_name, MAXFILENAME);
	strncpy (key.fieldname, oldname, MAXFIELDNAME);

	/*
	 * Locate the specific record entry and read it into the
	 * record buffer using index #5, the <filename,fieldname>
	 * index.
	 */

	if (__db_read_record (dict, 5, (void *) &key,
					IS_EQUAL, (char *) &record))
		return (db_errno = DB_BAD_DB), -1;

	/*
	 * Change the field name in the data dictionary entry to
	 * the new field name.
	 */

#ifdef	BSD
	bzero (record.f_fieldname, MAXFIELDNAME);
#else
	memset ((char *) record.f_fieldname, 0, MAXFIELDNAME);
#endif
	strncpy (record.f_fieldname, newname, MAXFIELDNAME);

	/*
	 * Update the record in the data dictionary.
	 */

	if (__db_update_record (dict, &record, 0))
		return (db_errno = DB_BAD_DB), -1;

	/*
	 * Update the entry in the schema for the currently open
	 * datebase.
	 */

	field = file->f_schema->s_fields + record.f_fieldno;
#ifdef	BSD
	bzero (field->f_name, MAXFIELDNAME);
#else
	memset ((char *) field->f_name, 0, MAXFIELDNAME);
#endif
	strncpy (field->f_name, newname, MAXFIELDNAME);

	return 0;
}

/*
 * db_rename_file - rename an existing database file
 */

int
db_rename_file (dbm, filenum, newname)
dbm_t	*dbm;
int	filenum;
char	*newname;
{
	file_t	*dict = &dbm->d_dictionary;
	file_t	*file;
	struct	{
		char	filename[MAXFILENAME];
	} key;
	dict_t	record;
	struct	stat	sb1, sb2;
	int	i;
	char	*cp;
	char	*oldname;
	char	oldpath[BUFSIZ];
	char	newpath[BUFSIZ];
	char	*dbpath;

	/*
	 * Validate the file number.
	 */

	if (filenum <= 0 || filenum > dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	oldname = file->f_file_name;

	/*
	 * Validate the new file name.  It must not exist in database
	 * at present.
	 */

	for (i = 1;i < dbm->d_file_count;i++)
		if (strncmp (dbm->d_files[i].f_file_name, newname,
				MAXFILENAME) == 0)
			return -1;

	/*
	 * Rename the primary index field
	 */

	sprintf (oldpath, ".%s", oldname);
	sprintf (newpath, ".%s", newname);
	if (db_rename_field (dbm, filenum, oldpath, newpath))
		return -1;

	/*
	 * Locate the physical files.  If they cannot be found, I will
	 * not be able to rename them in the end.
	 */

	if (fstat (file->f_key_fd, &sb1))
		return -1;

	if (! (dbpath = getenv ("DBPATH")))
		dbpath = ".";

	while (dbpath) {
		if (cp = strchr (dbpath, ':')) {
			strncpy (oldpath, dbpath, cp - dbpath);
			oldpath[cp - dbpath] = '\0';
		} else
			strcpy (oldpath, dbpath);

		strcat (oldpath, "/");
		strcat (oldpath, file->f_file_name);
		strcat (oldpath, ".idx");

		if (stat (oldpath, &sb2) == 0 &&
				sb1.st_ino == sb2.st_ino &&
				sb1.st_dev == sb2.st_dev)
			break;

		dbpath = cp;
	}
	if (! dbpath)
		return (db_errno = DB_NO_FILE), -1;

	/*
	 * Rewind the database dictionary
	 */

	dict->f_current_index = 1;
	if (isam_rewind (dict))
		return -1;

	/*
	 * Build the search key that will be used to locate the
	 * data dictionary entry for this particular file.
	 */

#ifdef	BSD
	bzero (&key, sizeof key);
#else
	memset ((char *) &key, 0, sizeof key);
#endif
	strncpy (key.filename, file->f_file_name, MAXFILENAME);

	/*
	 * Locate the specific record entry and read it into the
	 * record buffer using index #3, the <filename> index.
	 */

	while (__db_read_record (dict, 3, (void *) &key,
			IS_EQUAL, (char *) &record) == 0) {

		/*
		 * Change the file name in the data dictionary entry to
		 * the new file name.
		 */

#ifdef	BSD
		bzero (record.f_fieldfile, MAXFILENAME);
#else
		memset ((char *) record.f_fieldfile, 0, MAXFILENAME);
#endif
		strncpy (record.f_fieldfile, newname, MAXFILENAME);

		/*
		 * Update the record in the data dictionary.
		 */

		if (__db_update_record (dict, &record, 0))
			return (db_errno = DB_BAD_DB), -1;
	}

	/*
	 * Rename the physical files.
	 *
	 * NEEDSWORK: There are a zillion things that can go wrong
	 * here.  Figure them all out ;-)
	 */

	if (! cp)
		cp = dbpath + strlen (dbpath);

	sprintf (oldpath, "%.*s/%s.idx", cp - dbpath, dbpath, oldname);
	sprintf (newpath, "%.*s/%s.idx", cp - dbpath, dbpath, newname);
#ifdef	HAS_RENAME
	(void) rename (oldpath, newpath);
#else
	if (! link (oldpath, newpath))
		(void) unlink (oldpath);
#endif
	sprintf (oldpath, "%.*s/%s.dat", cp - dbpath, dbpath, oldname);
	sprintf (newpath, "%.*s/%s.dat", cp - dbpath, dbpath, newname);
#ifdef	HAS_RENAME
	(void) rename (oldpath, newpath);
#else
	if (! link (oldpath, newpath))
		(void) unlink (oldpath);
#endif

	/*
	 * Update the entry in the schema for the currently open
	 * datebase.
	 */

#ifdef	BSD
	bzero (file->f_file_name, MAXFILENAME);
#else
	memset ((char *) file->f_file_name, 0, MAXFILENAME);
#endif
	strncpy (file->f_file_name, newname, MAXFILENAME);

	return 0;
}
