/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#ifdef	MALLOC_H
#include <malloc.h>
#endif
#ifdef	STDLIB_H
#include <stdlib.h>
#endif
#include "is_types.h"
#include "is_proto.h"
#include "isam.h"

#ifdef BSD
#define memcpy(a, b, c) bcopy((b), (a), (c))
#else
#include <memory.h>
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)db_view.c	1.9	10:32:04	2/26/92";
#endif

#if !defined(MALLOC_H) && !defined(STDLIB_H)
extern	char	*malloc();
extern	void	free();
#endif

extern	int	db_errno;

#define MAX_VIEW_FIELDS		256	/* Descriptive view size limit */

struct	char_align {
	char	dummy;
	char	object;
};
struct	short_align {
	char	dummy;
	short	object;
};
struct	int_align {
	char	dummy;
	int	object;
};
struct	long_align {
	char	dummy;
	long	object;
};
struct	float_align {
	char	dummy;
	float	object;
};
struct	double_align {
	char	dummy;
	double	object;
};

struct	align {
	int	a_type;
	int	a_bound;
} aligns[] = {
	T_CHAR,		sizeof (struct char_align) - sizeof (char),
	T_SHORT,	sizeof (struct short_align) - sizeof (short),
	T_INT,		sizeof (struct int_align) - sizeof (int),
	T_LONG,		sizeof (struct long_align) - sizeof (long),
	T_DATE,		sizeof (struct long_align) - sizeof (long),
	T_SERIAL,	sizeof (struct long_align) - sizeof (long),
	T_FLOAT,	sizeof (struct float_align) - sizeof (float),
	T_DOUBLE,	sizeof (struct double_align) - sizeof (double),
	T_MONEY,	sizeof (struct double_align) - sizeof (double),
	0,		0
};

static int
__db_align (type, offset)
int	type;
int	offset;
{
	int	i;

	for (i = 0;aligns[i].a_type;i++) {
		if (aligns[i].a_type != type)
			continue;

		offset += aligns[i].a_bound - 1;
		offset &= ~(aligns[i].a_bound - 1);
		break;
	}
	return offset;
}

void
__db_view2buf (file, view, in, out)
file_t	*file;
view_t	*view;
char	*in;
char	*out;
{
	field_t	*field;
	int	i;

	for (i = 0;i < view->v_field_count;i++) {
		field = &file->f_schema->s_fields[view->v_fields[i]];
		memcpy (out + field->f_offset, in + view->v_offsets[i],
			field->f_size);
	}
}

void
__db_buf2view (file, view, in, out)
file_t	*file;
view_t	*view;
char	*in;
char	*out;
{
	int	i;
	field_t	*field;

#ifdef	BSD
	bzero (out, view->v_size);
#else
	memset (out, 0, view->v_size);
#endif /* BSD */
	for (i = 0;i < view->v_field_count;i++) {
		field = &file->f_schema->s_fields[view->v_fields[i]];
		memcpy (out + view->v_offsets[i], in + field->f_offset,
			field->f_size);
	}
}

/*
 * db_view - create a view into a file
 *
 * db_view() creates a data structure which is used to manipulate a
 * subset of the fields in a given file.  A pointer to the resulting
 * buffer is returned.
 */

VOIDP *
db_view (dbm, filenum, cnt, fields, packed)
dbm_t	*dbm;
int	filenum;
int	cnt;
short	*fields;
int	packed;
{
	file_t	*file;
	view_t	*view;
	field_t	*field;
	int	i, j;
	int	offset;

	/*
	 * Validate the parameters.  `filenum' must refer to a valid
	 * open file.  `cnt' must be greater than or equal to zero,
	 * and no field may be given more than once. `fields' is a list
	 * of data fields.  `packed' determines if the data objects are
	 * in consecutive memory locations, or if padding exists between
	 * each object as required by hardware constraints.
	 */

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), (VOIDP *) 0;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), (VOIDP *) 0;

	/*
	 * As a special case, if `cnt' is zero, free the current view.
	 */

	if (cnt == 0) {
		if (file->f_current_view) {
			free (file->f_current_view->v_fields);
			free (file->f_current_view->v_offsets);
			free (file->f_current_view->v_buffer);
			free (file->f_current_view);
			file->f_current_view = 0;
		}
		return (VOIDP *) file->f_data_buffer;
	}

	/*
	 * Scan through the list of fields and make certain they make
	 * sense.
	 */

	if (cnt < 0 || cnt > file->f_schema->s_field_count)
		return (db_errno = DB_INVALID), (VOIDP *) 0;

	for (i = 0;i < cnt;i++) {
		if (fields[i] < 0 || fields[i] >= file->f_schema->s_field_count)
			return (db_errno = DB_INVALID), (VOIDP *) 0;

		if (file->f_schema->s_fields[fields[i]].f_type == T_INDEX)
			return (db_errno = DB_INVALID), (VOIDP *) 0;

		for (j = 0;j < i;j++)
			if (fields[i] == fields[j])
				return (db_errno = DB_INVALID), (VOIDP *) 0;
	}

	/*
	 * Set up the structure to hold the new file view.  If there
	 * was already a view, free the storage allocated for it.
	 * Otherwise, allocate space to hold the parts.
	 */

	if (file->f_current_view) {
		free (file->f_current_view->v_fields);
		free (file->f_current_view->v_offsets);
		free (file->f_current_view->v_buffer);
	} else {
		if (! (file->f_current_view = (view_t *)
				malloc (sizeof *file->f_current_view)))
			return (db_errno = DB_NOMEM), (VOIDP *) 0;
	}

	/*
	 * Build the new view from the supplied parts.  A list of field
	 * numbers and their offsets into the record is created from the
	 * input arguments.
	 */

	view = file->f_current_view;

	if (! (view->v_offsets = (short *) malloc (cnt * sizeof (short)))) {
		db_errno = DB_NOMEM;
		goto fail;
	}
	if (! (view->v_fields = (short *) malloc (cnt * sizeof (short)))) {
		db_errno = DB_NOMEM;
		goto fail;
	}
	for (offset = 0, i = 0;i < cnt;i++) {
		field = &file->f_schema->s_fields[fields[i]];

		view->v_fields[i] = fields[i];
		if (! packed)
			offset = __db_align (field->f_type, offset);

		view->v_offsets[i] = offset;
		offset += field->f_size;
	}

	/*
	 * If the view is unpacked, the size must be rounded up to
	 * the largest sized object.  For most systems this will be
	 * a T_DOUBLE object.
	 */

	if (! packed)
		offset = __db_align (T_DOUBLE, offset);

	/*
	 * Allocate a buffer large enough to hold the parts of a record
	 * that belong to the file view.
	 */

	view->v_size = offset;
	if (! (view->v_buffer = malloc (offset))) {
		db_errno = DB_NOMEM;
		goto fail;
	}
	view->v_packed = (packed != 0);
	view->v_field_count = cnt;
	return (VOIDP *) view->v_buffer;

	/*
	 * After the view has been allocated, all failures lead to
	 * this point.
	 */

fail:
	if (view->v_offsets)
		free (view->v_offsets);

	if (view->v_fields)
		free (view->v_fields);

	if (view->v_buffer)
		free (view->v_buffer);

	free (view);
	file->f_current_view = 0;
	return (VOIDP *) 0;
}

int
db_read_view (dbm, filenum, data, match)
dbm_t	*dbm;
int	filenum;
VOIDP	*data;
int	match;
{
	file_t	*file;
	view_t	*view;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	if (db_read_record (dbm, filenum, data, match))
		return -1;

	if ((view = file->f_current_view) == 0)
		return 0;

	__db_buf2view (file, view, file->f_data_buffer, view->v_buffer);
	return 0;
}

int
db_write_view (dbm, filenum, buffer)
dbm_t	*dbm;
int	filenum;
VOIDP	*buffer;
{
	file_t	*file;
	view_t	*view;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

#ifdef	BSD
	bzero (file->f_data_buffer, file->f_record_size);
#else
	memset (file->f_data_buffer, 0, file->f_record_size);
#endif /* BSD */
	if (view = file->f_current_view) {
		__db_view2buf (file, view, buffer, file->f_data_buffer);
		return db_write_record (dbm, filenum, file->f_data_buffer);
	} else
		return db_write_record (dbm, filenum, buffer);
}

/*
 * db_update_view - update the fields in the current file view
 *
 * db_update_view() uses the currently defined file view and uses it to
 * update the fields in the current record.  It is identical to
 * db_update_record except that all of the fields are not modified.
 */

int
db_update_view (dbm, filenum, buffer)
dbm_t	*dbm;
int	filenum;
VOIDP	*buffer;
{
	file_t	*file;
	char	spare[MAXRECSIZE];
	int	was_locked;
	long	ticket;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	if (file->f_current_record == -1)
		return (db_errno = DB_NO_RECORD), -1;

	was_locked = file->f_locked;
	if (was_locked != F_WRLCK)
		if ((ticket = isam_set_ticket (file)) == -1)
			return -1;

#ifdef	BSD
	bzero (spare, sizeof spare);
#else
	memset (spare, 0, sizeof spare);
#endif /* BSD */

	/* 
	 * Read the old record data into a spare buffer and overlay
	 * the view onto the existing data.  Then update THAT record.
	 */

	if (isam_read_data (file, file->f_current_record, spare))
		goto fail;

	__db_view2buf (file, file->f_current_view, buffer, spare);
	if (db_update_record (dbm, filenum, spare))
		goto fail;

	if (was_locked != F_WRLCK) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return 0;

fail:
	if (was_locked != F_WRLCK) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return -1;
}

/*
 * db_desc_view - describe a new view
 *
 * db_desc_view() describes a file view by field name and type.  This allows
 * the application to request specific fields and sizes when creating new
 * views.
 */

VOIDP *
db_desc_view (dbm, filenum, desc, packed)
dbm_t	*dbm;
int	filenum;
desc_t	*desc;
int	packed;
{
	int	i;
	int	fieldnum;
	short	fields[MAX_VIEW_FIELDS];
	schema_t *schema;
	field_t	*field;

	/*
	 * Validate the filenum parameter.
	 */

	if (filenum < 0 || filenum > dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), (VOIDP *) 0;

	/*
	 * Check the view description.  It must be less than MAX_VIEW_FIELDS
	 * entries long.  Also, if there is no description, remove the view.
	 */

	if (desc) {

		/*
		 * This is a new view - check it for size
		 */

		for (i = 0;i < MAX_VIEW_FIELDS && desc[i].desc_name;i++)
			;

		if (i == MAX_VIEW_FIELDS)
			return (db_errno = DB_TOO_MANY), (VOIDP *) 0;
	} else {

		/*
		 * Remove any existing view
		 */

		return db_view (dbm, filenum, 0, (short *) 0, 0);
	}

	/* 
	 * Look for the named fields in the shema and build the
	 * list of fields to be passed to db_view.
	 */

	schema = dbm->d_files[filenum].f_schema;

	for (i = 0;desc[i].desc_name != (char *) 0;i++) {
		if ((fieldnum = db_open_field (dbm,
				filenum, desc[i].desc_name)) == -1)
			return (VOIDP *) 0;

		field = &schema->s_fields[fieldnum];
		if (field->f_type != desc[i].desc_type)
			return (db_errno = DB_BAD_DESC), (VOIDP *) 0;

		if (desc[i].desc_type == T_CHAR &&
				field->f_size != desc[i].desc_size)
			return (db_errno = DB_BAD_DESC), (VOIDP *) 0;

		fields[i] = fieldnum;
	}

	/*
	 * All of the fields for the view have been located and
	 * verified.  Now we call db_view() to turn our list of field
	 * names into a file view.
	 */

	return db_view (dbm, filenum, i, fields, packed);
}
