/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#ifndef	lint
static	char	sccsid[] = "@(#)dbscomp.c	1.6	10:33:41	2/26/92";
#endif

#include <sys/types.h>
#include <stdio.h>
#ifdef	MALLOC_H
#include <malloc.h>
#endif
#include "is_lex.h"
#include "isam.h"
#include "is_types.h"
#include "is_proto.h"
#include "patchlevel.h"

#ifdef BSD
#include <strings.h>
#define strchr index
#define memcpy(a, b, c) bcopy((b), (a), (c))
#else
#include <string.h>
#include <memory.h>
#endif /* BSD */

extern	char	yytext[];
extern	int	yylineno;
#ifndef	MALLOC_H
extern	char	*malloc();
extern	char	*realloc();
#endif

extern	int	db_errno;

find_variable (display, field)
bdisplay_t *display;
bfield_t *field;
{
	char	*line;
	char	*cp;
	int	len;
	int	i;

	len = strlen (field->bf_name);

	field->bf_y = 0;
	for (field->bf_y = 0, line = display->bd_background;
			*line;
			line += strlen (line) + 1, field->bf_y++) {
		for (cp = strchr (line, '[');cp;cp = strchr (cp, '[')) {
			cp++;
			if (strncmp (cp, field->bf_name, len) != 0)
				continue;

			if (strchr (" \t]", cp[len]))
				break;
		}
		if (cp)
			break;
	}
	for (field->bf_x = 0, i = 0;line + i != cp;i++) {
		if (line[i] == '\t') {
			field->bf_x += 8;
			field->bf_x &= ~7;
		} else
			field->bf_x++;
	}
}

line_length (line)
char	*line;
{
	int	i;

	for (i = 0;*line;line++) {
		if (*line != '\t')
			i++;
		else {
			i += 8;
			i &= ~7;
		}
	}
	return i;
}

main (argc, argv)
int	argc;
char	**argv;
{
	char	buf[BUFSIZ];
	char	scr_name[BUFSIZ];
	char	dummy[4096];
	char	*cp;
	int	token;
	int	i;
	int	errors = 0;
	bscreen_t *screen;
	bdisplay_t *display, *new_display;
	bfield_t *field;
	FILE	*scr;

	/*
	 * See if the user is asking for the command version number
	 * and display it
	 */

	if (argc > 2 && strcmp (argv[1], "-V") == 0) {
		fprintf (stderr, "dbscreen: version %s\n", VERSION);
		exit (0);
	}

	/*
	 * Check the command line syntax.  The file name must be given
	 * on the command line so we know what to name the output.  That
	 * file will be opened as stdin.
	 */

	if (argc != 2) {
		fprintf (stderr, "usage: dbscreen [ -V ] screen\n");
		exit (1);
	}
	if (freopen (argv[1], "r", stdin) == 0) {
		sprintf (buf, "dbscreen: can't open %s", argv[1]);
		perror (buf);
		exit (1);
	}

	/*
	 * Figure out the name of the output file and create it.
	 */

	strcpy (scr_name, argv[1]);
	strcat (scr_name, ".scr");
	if ((scr = fopen (scr_name, "w")) == 0) {
		sprintf (buf, "dbscreen: can't open %s", scr_name);
		perror (buf);
		exit (1);
	}

	/*
	 * Get and initialize the screen structure for this file
	 */

	if ((screen = (bscreen_t *) malloc (sizeof *screen)) == 0) {
		perror ("dbscreen: out of memory");
		goto fail;
	}
#ifdef BSD
	bzero (screen, sizeof *screen);
#else
	memset (screen, 0, sizeof *screen);
#endif /* BSD */
	screen->bs_screen_cnt = 0;
	screen->bs_size = sizeof (*screen) - sizeof (*display);

	/*
	 * Get the name of the database that this screen is for
	 */

	if ((token = yylex ()) != L_DATABASE) {
		fprintf (stderr, "line %d: expected DATABASE, got '%s'\n",
			yylineno, yytext);
		goto fail;
	}
	if (yylex () != L_NAME) {
		fprintf (stderr, "line %d: expected <name>, got '%s'\n",
			yylineno, yytext);
		goto fail;
	}
	strncpy (screen->bs_database, yytext, MAXFILENAME);

	/*
	 * Read all the screen definitions for this file
	 */

	for (token = yylex ();token == L_SCREEN && errors < 10;) {
		if ((display = (bdisplay_t *) malloc (sizeof *display)) == 0) {
			perror ("dbscreen: out of memory");
			goto fail;
		}
#ifdef BSD
		bzero (display, sizeof *display);
#else
		memset (display, 0, sizeof *display);
#endif /* BSD */
		cp = display->bd_background;
		*cp = '\0';

		if (fgets (buf, BUFSIZ, stdin) == 0) {
			yylineno++;
			fprintf (stderr, "line %d: expected '(', got EOF\n",
				yylineno);
			goto fail;
		}
		yylineno++;
		*(strchr (buf, '\n')) = '\0';

		/*
		 * The '(' on a line by itself starts the screen description.
		 * There should be one on the line right here.
		 */

		if (strcmp (buf, "(") != 0) {
			fprintf (stderr, "line %d: expected '(', got '%s'\n",
				yylineno, buf);
			goto fail;
		}

		/*
		 * Read everything between the two '(' ')' delimiters into
		 * the screen image part of the display.  The parts will be
		 * located later.
		 */

		while (1) {
			if (fgets (buf, BUFSIZ, stdin) == 0) {
				yylineno++;
				fprintf (stderr, "line %d: unexpected EOF\n",
					yylineno);
				goto fail;
			}
			yylineno++;
			*(strchr (buf, '\n')) = '\0';
			if (buf[0] == ')' && buf[1] == '\0')
				break;

			if (! buf[0])
				strcpy (buf, " ");

			if (line_length (buf) > display->bd_width)
				display->bd_width = line_length (buf);

			display->bd_height++;

			strcpy (cp, buf);
			cp += strlen (cp) + 1;
		}
		*cp++ = '\0';

		display->bd_size = sizeof (*display) - sizeof (*field);
		token = yylex ();

		/*
		 * Parse the field and file joining descriptions.  Fields
		 * listed in the screen description must be associated with
		 * some field in a database file.  If more than one file is
		 * named in the screen, a collection of relational joins
		 * must be described.
		 */

		for (;(token == L_NAME || token == L_FILET) && errors < 10;
				token = yylex ()) {
			if (token == L_NAME) {

				/*
				 * Add a new screen variable
				 */

				for (i = 0;i < display->bd_field_cnt;i++) {
					if (! strcmp (yytext,
						display->bd_fields[i].bf_name))
						break;
				}

				/*
				 * Make certain it didn't already exist.
				 * If it did, report and log the error.
				 * If it didn't, allocate space.
				 */

				if (i != display->bd_field_cnt) {
					fprintf (stderr,
						"line %d: redefining %s\n",
						yylineno, yytext);
					errors++;
				} else
					display->bd_size += sizeof (*field);

				if (! (display = (bdisplay_t *)
					realloc (display, display->bd_size))) {
					perror ("dbscreen: out of memory");
					goto fail;
				}

				/*
				 * Initialize the new field
				 */

				field = display->bd_fields + i;
#ifdef BSD
				bzero (field,  sizeof *field);
#else
				memset (field, 0,  sizeof *field);
#endif /* BSD */
				if (i == display->bd_field_cnt)
					display->bd_field_cnt++;

				strncpy (field->bf_name, yytext, MAXFIELDNAME);

				/* 
				 * See if this is a read-only field
				 */

				if ((token = yylex ()) == L_READONLY) {
					field->bf_readonly = 1;
					token = yylex ();
				} else
					field->bf_readonly = 0;

				/*
				 * Get the file name that the field is
				 * a member of 
				 */

				if (token != L_FILET) {
					fprintf (stderr,
						"line %d: expected `FILE'\n",
						yylineno, yytext);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				if (yylex () != L_NAME) {
					fprintf (stderr,
						"line %d: expected <name>\n",
						yylineno, yytext);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				strncpy (field->bf_file, yytext, MAXFILENAME);

				/*
				 * Get the name of the field within the
				 * database file that is to be used for
				 * the current variable.
				 */

				if (yylex() != L_FIELD) {
					fprintf (stderr,
						"line %d: expected `FIELD'\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				if (yylex () != L_NAME) {
					fprintf (stderr,
						"line %d: expected <name>\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				strncpy (field->bf_field, yytext, MAXFIELDNAME);

				/*
				 * Now go look for the variable in the screen
				 * image buffer.  It should be in there
				 * somewheres ...
				 */

				if (! find_variable (display, field)) {
					fprintf (stderr,
						"line %d: %s not found\n",
						yylineno, field->bf_name);
					errors++;
				}
			} else {

				/*
				 * Add a new file joining definition
				 */

				display->bd_size += sizeof (*field);
				if (! (display = (bdisplay_t *)
					realloc (display, display->bd_size))) {
					perror ("dbscreen: out of memory");
					goto fail;
				}

				/* 
				 * Joins live in file definitions, but don't
				 * have names or anything else like that.
				 */

				field = display->bd_fields +
					display->bd_field_cnt;
#ifdef BSD
				bzero (field,  sizeof *field);
#else
				memset (field, 0,  sizeof *field);
#endif /* BSD */
				field->bf_join = 1;
				display->bd_field_cnt++;

				/*
				 * Get the name of the left-hand file.  It
				 * will be used as the key into the right-hand
				 * file.
				 */

				if (token != L_FILET) {
					fprintf (stderr,
						"line %d: expected `FILE'\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				if (yylex () != L_NAME) {
					fprintf (stderr,
						"line %d: expected <name>\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				strncpy (field->bf_join_file, yytext,
					MAXFILENAME);

				/*
				 * Get the name of the left-hand field.  This
				 * is the actual field to use for the search.
				 */

				if (yylex() != L_FIELD) {
					fprintf (stderr,
						"line %d: expected `FIELD'\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				if (yylex () != L_NAME) {
					fprintf (stderr,
						"line %d: expected <name>\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					token = yylex ();
					continue;
				}
				strncpy (field->bf_join_field, yytext,
					MAXFIELDNAME);

				/*
				 * Get the "joining" reserved word.  It starts
				 * the right-hand side of the join clause.
				 */

				if (yylex () != L_JOINING) {
					fprintf (stderr,
						"line %d: expected `JOINING'\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}

				/*
				 * Get the right-hand file.  This file will be
				 * searched to find the matching record.
				 */

				if (yylex () != L_FILET) {
					fprintf (stderr,
						"line %d: expected `FILE'\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				if (yylex () != L_NAME) {
					fprintf (stderr,
						"line %d: expected <name>\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					yylineno++;
					continue;
				}
				strncpy (field->bf_file, yytext, MAXFILENAME);

				/*
				 * Get the right-hand field.  This should be
				 * an index field so the search is efficient.
				 */

				if (yylex() != L_FIELD) {
					fprintf (stderr,
						"line %d: expected `FIELD'\n",
						yylineno);
					errors++;
					fgets (buf, sizeof buf, stdin);
					continue;
				}
				if (yylex () != L_NAME) {
					fprintf (stderr,
						"expected <name>, got `%s'\n",
						yytext);
					exit (1);
				}
				strncpy (field->bf_field, yytext, MAXFIELDNAME);
			}
		}

		/*
		 * Allocate the memory to stuff this new thing into
		 */

		if ((screen = (bscreen_t *) realloc (screen,
				screen->bs_size + display->bd_size)) == 0) {
			perror ("dbscreen: out of memory");
			exit (1);
		}
		memcpy ((char *) screen + screen->bs_size, display,
				display->bd_size);
		screen->bs_size += display->bd_size;
		free (display);
		screen->bs_screen_cnt++;
	}
	if (errors)
		goto fail;

	/*
	 * Write the entire file in one go
	 */

	if (fwrite ((char *) screen, screen->bs_size, 1, scr) != 1 ||
			fflush (scr) == -1) {
		sprintf (buf, "dbscreen: can't write %s", scr_name);
		perror (buf);
		exit (1);
	}
	fclose (scr);
	exit (0);

fail:
	unlink (scr_name);
	exit (1);
}
