/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include "is_types.h"

#ifndef lint
static	char	sccsid[] = "@(#)is_ticket.c	1.5	11:49:21	1/29/92";
#endif

#ifdef	NO_LOCKING
static	long	lock_cnt = 1;
#endif

extern	int	db_errno;
extern	int	errno;

long
isam_get_ticket (file)
file_t	*file;
{
#ifdef	NO_LOCKING
	if (! file->f_locked)
		file->f_locked = 1;

	return lock_cnt;
#else
	struct	flock	flock;
	long	ticket;
	int	was_locked = file->f_locked;

	if (! was_locked) {
		flock.l_type = F_RDLCK;
		flock.l_whence = 0;
		flock.l_start = sizeof (off_t);
		flock.l_len = sizeof (long);

		if (fcntl (file->f_key_fd, F_SETLK, &flock)) {
			if (errno == EAGAIN)
				return (db_errno = DB_IS_LOCKED), -1;

			return (db_errno = DB_EIO), -1;
		}
	} else {
		flock.l_type = F_RDLCK;
		flock.l_whence = 0;
		flock.l_start = sizeof (off_t);
		flock.l_len = sizeof (long);

		if (fcntl (file->f_key_fd, F_GETLK, &flock))
			return (db_errno = DB_EIO), -1;

		if (flock.l_type != F_UNLCK)
			return (db_errno = DB_EIO), -1;
	}
	if (lseek (file->f_key_fd, (long) sizeof (off_t), 0) == -1) {
		db_errno = DB_EIO;
		goto fail;
	}
	if (read (file->f_key_fd, &ticket, sizeof ticket) != sizeof ticket) {
		db_errno = DB_EIO;
		goto fail;
	}
	if (! was_locked)
		file->f_locked = F_RDLCK;

	return ticket;

fail:
	if (! was_locked) {
		flock.l_type = F_UNLCK;
		(void) fcntl (file->f_key_fd, F_SETLK, &flock);
		file->f_locked = 0;
	}
	return -1;
#endif
}

long
isam_get_ticket_wait (file)
file_t	*file;
{
#ifdef	NO_LOCKING
	if (! file->f_locked)
		file->f_locked = 1;

	return lock_cnt;
#else
	struct	flock	flock;
	long	ticket;
	int	was_locked = file->f_locked;

	if (! was_locked) {
		flock.l_type = F_RDLCK;
		flock.l_whence = 0;
		flock.l_start = sizeof (off_t);
		flock.l_len = sizeof (long);

		if (fcntl (file->f_key_fd, F_SETLKW, &flock))
			return (db_errno = DB_EIO), -1;
	} else {
		flock.l_type = F_RDLCK;
		flock.l_whence = 0;
		flock.l_start = sizeof (off_t);
		flock.l_len = sizeof (long);

		if (fcntl (file->f_key_fd, F_GETLK, &flock))
			return (db_errno = DB_EIO), -1;

		if (flock.l_type != F_UNLCK)
			return (db_errno = DB_EIO), -1;
	}
	if (lseek (file->f_key_fd, (long) sizeof (off_t), 0) == -1) {
		db_errno = DB_EIO;
		goto fail;
	}
	if (read (file->f_key_fd, &ticket, sizeof ticket) != sizeof ticket) {
		db_errno = DB_EIO;
		goto fail;
	}
	if (! was_locked)
		file->f_locked = F_RDLCK;

	return ticket;

fail:
	if (! was_locked) {
		flock.l_type = F_UNLCK;
		(void) fcntl (file->f_key_fd, F_SETLK, &flock);
		file->f_locked = 0;
	}
	return -1;
#endif
}

long
isam_set_ticket (file)
file_t	*file;
{
#ifdef	NO_LOCKING
	if (! file->f_locked)
		file->f_locked = 2;

	return ++lock_cnt;
#else
	struct	flock	flock;
	long	ticket;

	flock.l_type = F_WRLCK;
	flock.l_whence = 0;
	flock.l_start = sizeof (off_t);
	flock.l_len = sizeof (long);

	if (fcntl (file->f_key_fd, F_SETLK, &flock)) {
		if (errno == EAGAIN)
			return (db_errno = DB_IS_LOCKED), -1;

		return (db_errno = DB_EIO), -1;
	}
	if (lseek (file->f_key_fd, (long) sizeof (off_t), 0) == -1) {
		db_errno = DB_EIO;
		goto fail;
	}
	if (read (file->f_key_fd, &ticket, sizeof ticket) != sizeof ticket) {
		db_errno = DB_EIO;
		goto fail;
	}
	ticket++;
	if (ticket == 0 || ticket == -1)
		ticket = 1;

	if (lseek (file->f_key_fd, (long) sizeof (off_t), 0) == -1) {
		db_errno = DB_EIO;
		goto fail;
	}
	if (write (file->f_key_fd, &ticket, sizeof ticket) != sizeof ticket) {
		db_errno = DB_EIO;
		goto fail;
	}
	file->f_locked = F_WRLCK;
	return ticket;

fail:
	flock.l_type = F_UNLCK;
	(void) fcntl (file->f_key_fd, F_SETLK, &flock);
	file->f_locked = 0;
	return -1;
#endif
}

long
isam_set_ticket_wait (file)
file_t	*file;
{
#ifdef	NO_LOCKING
	if (! file->f_locked)
		file->f_locked = 2;

	return ++lock_cnt;
#else
	struct	flock	flock;
	long	ticket;

	flock.l_type = F_WRLCK;
	flock.l_whence = 0;
	flock.l_start = sizeof (off_t);
	flock.l_len = sizeof (long);

	if (fcntl (file->f_key_fd, F_SETLKW, &flock))
		return (db_errno = DB_EIO), -1;

	if (lseek (file->f_key_fd, (long) sizeof (off_t), 0) == -1) {
		db_errno = DB_EIO;
		goto fail;
	}
	if (read (file->f_key_fd, &ticket, sizeof ticket) != sizeof ticket) {
		db_errno = DB_EIO;
		goto fail;
	}
	ticket++;
	if (ticket == 0 || ticket == -1)
		ticket = 1;

	if (lseek (file->f_key_fd, (long) sizeof (off_t), 0) == -1) {
		db_errno = DB_EIO;
		goto fail;
	}
	if (write (file->f_key_fd, &ticket, sizeof ticket) != sizeof ticket) {
		db_errno = DB_EIO;
		return -1;
	}
	file->f_locked = F_WRLCK;
	return ticket;

fail:
	flock.l_type = F_UNLCK;
	file->f_locked = 0;
	(void) fcntl (file->f_key_fd, F_SETLK, &flock);
	return -1;
#endif
}

int
isam_end_ticket (file)
file_t	*file;
{
#ifdef	NO_LOCKING
	file->f_locked = 0;
#else
	struct	flock	flock;

	flock.l_type = F_UNLCK;
	flock.l_whence = 0;
	flock.l_start = sizeof (off_t);
	flock.l_len = sizeof (long);
	if (fcntl (file->f_key_fd, F_SETLK, &flock)) {
		file->f_locked = 0;
		return (db_errno = DB_EIO), -1;
	}
	file->f_locked = 0;
	return 0;
#endif
}
